"""
    Created on 2012-01-20
    @author: jldupont
"""
import sys, os
import logging
import copy
from time import sleep

from tools_logging import info_dump
from tools_mod     import call
from tools_s3      import keys_to_dict, compute_changes, bucket_status, json_string, key_object_to_dict

def stdout(s):
    sys.stdout.write(s+"\n")


def output_json(bucket_name, prefix, keys, changes, additions, substractions):
    o={
       "bucket_name": bucket_name
       ,"prefix": prefix
       }
    
    rkeys={}
    for key in keys:
        ko=key_object_to_dict(key)
        rkeys.update(ko)
        
    o["keys"]=rkeys
    o["changes"]=changes
    o["additions"]=additions
    o["substractions"]=substractions
    
    stdout(json_string(o))


def run(args):
    """
    args.bucket_name
    args.module_name
    """
    cdict={}
    
    if args.enable_debug:
        logger=logging.getLogger()
        logger.setLevel(logging.DEBUG)
    
    bucket_name=args.bucket_name.strip()
    module_name=args.module_name
    polling=args.polling_interval
    enable_json=args.enable_json
    propagate_error=args.propagate_error
    prefix=args.prefix
    always=args.always

    if module_name is not None:
        module_name=module_name.strip()
        
    info_dump(args._get_kwargs(), 20)
    
    ppid=os.getppid()
    logging.info("Process pid: %s" % os.getpid())
    logging.info("Parent pid: %s" % ppid)
    logging.info("Starting loop...")
    while True:
        if os.getppid()!=ppid:
            logging.warning("Parent terminated... exiting")
            break

        logging.debug("Performing bucket 'get_all_keys'")
                
        try:
            code, data=bucket_status(bucket_name, prefix)
        except Exception,e:
            logging.error("Can't reach S3")
            code="error"
            if propagate_error:
                stdout('''{"error": "Can't reach S3"}''')
            
        if code.startswith("ok"):
            try:
                ndict=keys_to_dict(data)
                changes, additions, substractions=compute_changes(cdict, ndict)
                
                if len(changes)>0 or len(additions)>0 or len(substractions)>0 or always:
                    try:
                        if module_name is not None:
                            call(module_name, "run", bucket_name, prefix, keys=ndict, changes=changes, additions=additions, substractions=substractions)
                            logging.debug("success making the 'run' call")                        
                    except Exception, e:
                        logging.error("calling 'run' function of module '%s': %s" %(module_name, str(e)))
                    
                            
                    try:
                        if enable_json:
                            output_json(bucket_name, prefix, data, changes, additions, substractions)
                            logging.debug("success with output of json to stdout")
                            
                    except Exception, e:
                        logging.error("generating json output to stdout: %s" % str(e))
                        
                cdict=dict(ndict)
                
            except:
                logging.error("checking changes in bucket '%s', prefix '%s'" % (bucket_name, prefix))
        
        logging.debug("...sleeping for %s seconds" % polling)
        sleep(polling)
    
    



"""
    Created on 2012-01-20
    @author: jldupont
"""
import sys, logging
from time import sleep

from tools_s3    import json_string, get_all_keys
from tools_os    import resolve_path, safe_path_exists
from tools_os    import can_write, check_can_write_paths, check_exist_paths
from tools_sys   import retry
from tools_func  import transition_manager
import boto
from boto.s3.key import Key as S3Key

from pyfnc import patterned, pattern, partial, dic, coroutine

def stdout(s):
    sys.stdout.write(s+"\n")


def log(msg, path):
    logging.info(msg % path)
    


def run(bucket_name=None, prefix=None,
        path_dest=None, path_check=None,
        dont_dl=None, just_keys=None,
        num_files=5, propagate_error=False,  polling_interval=None):
    
    if path_check is not None:
        code, path_check=resolve_path(path_check)
        if not code.startswith("ok"):
            logging.warning("path_check '%s' might be in error..." % path_check)
    
    ### VALIDATE PARAMETERS
    
    code, path_dest=resolve_path(path_dest)
    if not code.startswith("ok"):
        raise Exception("Invalid destination path: %s" % path_dest)
    
    code, _path=can_write(path_dest)
    if not code.startswith("ok"):
        raise Exception("Can't write to destination path: %s" % path_dest)
    
    logging.info("* Can write to destination path")
    
    try:
        conn = boto.connect_s3()
    except:
        ## not much we can do
        ## but actually no remote calls are made
        ## at this point so it should be highly improbable
        raise Exception("Can't 'connect' to S3")
    
    ###################### BUCKET
    logging.info("Getting/creating bucket (unlimited retries with backoff)")
    def _get_create_bucket():
        return conn.create_bucket(bucket_name)
              
    bucket=retry(_get_create_bucket)
    logging.info("Got bucket")
    #############################

    

    ctx=dic({
         "bucket":     bucket
         ,"prefix":    prefix
         ,"path_dest": path_dest
         ,"just_keys": just_keys
         ,"dont_dl":   dont_dl
         ,"num_files": num_files
         ,"cp": {
                 "up":    partial(log, "Check path OK:  %s", path_check)
                 ,"down": partial(log, "Check path !OK: %s", path_check) 
                 }
         })

    ctx["tm"]=transition_manager(ctx)
    ctx["mk"]=manager_keys(ctx)
    
    
    logging.info("Starting loop...")
    
    while True:
        
        #################################################
        try:
            _code, path_exists=safe_path_exists(path_check)            
            maybe_process(ctx, path_check, path_exists)
            
        except Exception, e:
            logging.error(str(e))

        #####################################################
        logging.debug("...sleeping for %s seconds" % polling_interval)
        sleep(polling_interval)
    
    
@pattern(dic, None, any)
def maybe_process_1(ctx, _1, _2):
    process(ctx)

@pattern(dic, any, False)
def maybe_process_2(ctx, _1, _2):
    """ nothing to do has check_path doesn't exist """
    ctx["tm"].send(("cp", False))
    
@pattern(dic, any, True)
def maybe_process_3(ctx, _1, _2):
    ctx["tm"].send(("cp", True))
    process(ctx)

@patterned
def maybe_process(ctx, path_check, path_check_exists): pass


def process(ctx):
    """
    Really process from here :)
    """
    bucket=ctx["bucket"]
    prefix=ctx["prefix"]
    path_dest=ctx["path_dest"]
    
    code, maybe_keys=get_all_keys(bucket, prefix)
    if not code.startswith("ok"):
        e=maybe_keys
        logging.warning("Error retrieving keys: %s" % str(e))
        return

    ## Check if it already exists 
    results_exists=check_exist_paths(path_dest, maybe_keys)

    def remove_exists(entry):
        code, (_path_fragment, exists)=entry
        if not code.startswith("ok"):
            return False ## filter-out
        return not exists
    
    potential_todo=filter(remove_exists, results_exists)

    def keep_only_keys(entry):
        _, (key, state)=entry
        return key
    
    potential_todo_keys=map(keep_only_keys, potential_todo)

    ## Check if 'key' can be written to
    results=check_can_write_paths(path_dest, potential_todo_keys)
    
    def keep(which):
        
        def switch(status):
            return status if which=="writable" else not status
        
        def _(entry):
            code, (_fragment, status)=entry
            if not code.startswith("ok"):
                return switch(False)
            
            return switch(status)
        return _
            
    good_key_entries=filter(keep("writable"), results)
    bad_key_entries=filter(keep("!writable"), results)

    ## Report bad keys to manager_keys
    ctx["mk"].send(bad_key_entries)
        
    process2(ctx, ctx["just_keys"], good_key_entries)
    
    
@pattern(dic, True, list)
def process2_justkeys(ctx, _, keys):
    """
    """
    print keys
    
@pattern(dic, False, list)
def process2_dl(ctx, _, keys):
    """
    """
    print keys
    
@patterned
def process2(ctx, just_keys, keys): pass


    

@coroutine
def manager_keys(ctx):
    
    while True:
        entries=(yield)
        


        
        



"""
    Various module level functions
    
    Created on 2012-01-20
    @author: jldupont
"""
import importlib

def _echo(st):
    return st

def prepare_callable(module_name, function_name):
    
    if module_name is None or function_name is None:
        raise Exception("Invalid module/function name")
    
    try:
        mod=importlib.import_module(module_name)
    except:
        raise Exception("Can't import module '%s'" % module_name)
        
    try:
        fnc=getattr(mod, function_name)
    except:
        raise Exception("Module '%s' doesn't have a 'run' function" % mod)
    
    if not callable(fnc):
        raise Exception("Can't call 'run' function of callable '%S'" % module_name)
    
    return (mod, fnc)

def call(module_name, function_name, *pargs, **kargs):
    """
    >>> call("tools_mod", "_echo", "hello!")
    'hello!'
    """
    try:
        mod=importlib.import_module(module_name)
    except:
        raise Exception("Can't import module '%s'" % module_name)
        
    try:
        fnc=getattr(mod, function_name)
    except:
        raise Exception("Module '%s' doesn't have a 'run' function" % mod)
    
    if not callable(fnc):
        raise Exception("Can't call 'run' function of callable '%S'" % module_name)
    
    return fnc(*pargs, **kargs)
    

if __name__=="__main__":
    import doctest
    doctest.testmod()
    


"""
    Created on 2012-01-20
    @author: jldupont
"""
import logging

import boto
from boto.s3.key import Key as S3Key

def run(bucket_name=None, bucket_prefix=None, alternate_format=False):
    
    try:
        conn = boto.connect_s3()
    except:
        ## not much we can do
        ## but actually no remote calls are made
        ## at this point so it should be highly improbable
        raise Exception("Can't 'connect' to S3")
    
    try:
        bucket=conn.create_bucket(bucket_name)
        logging.info("Got bucket '%s'" % bucket_name)        
    except:
        raise Exception("Can't get bucket '%s'" % bucket_name)
    
    """
        s3://bucket_name/bucket_prefix/key
        s3://bucket_name/key
        bucket_name /t bucket_prefix/key
    """
    
    if alternate_format:
        _base_format="%s\t" % bucket_name
    else:
        _base_format="s3://%s/" % bucket_name
    
    if bucket_prefix is not None:
        liste=bucket.list(prefix=bucket_prefix)
    else:
        liste=bucket.list()
        
    
    for key in liste:
        if not key.name.endswith("/"):
            print "%s%s" % (_base_format, key.name)



"""
    Created on 2012-01-19
    @author: jldupont
"""
import os

def build_topic_arn(conn, topic):
    account_id=os.environ["AWS_ACCOUNT_ID"]
    return 'arn:aws:sns:%s:%s:%s' % (conn.region.name, account_id, topic)



'''
    Simple Key/Value store organized as a "journal"
    
    * 'insert' timestamped KV pairs
    * retrieve list of timestamped KV pairs 

    Created on 2011-06-20
    @author: jldupont
'''
import os
import uuid
import boto
from jldaws.tools_date import iso_now_and_then
from jldaws.exceptions import SDB_Connect, SDB_CreateDomain, SDB_Access, SDB_CreateItem


class SimpleKV():
    
    def __init__(self, aws_access_key=None, aws_secret_key=None):
        self.aws_access_key=aws_access_key or os.environ.get("AWS_ACCESS_KEY_ID", None)
        self.aws_secret_key=aws_secret_key or os.environ.get("AWS_SECRET_ACCESS_KEY", None)
        
    def connect(self):
        """
        Connect to AWS SDB
        
        Use this method first before attempting any other
        
        @raise SDB_Connect:  
        """
        try:        
            if self.aws_access_key is None:
                self.sdbconn=boto.connect_sdb()
            else:
                self.sdbconn=boto.connect_sdb(self.aws_access_key, self.aws_secret_key)
        except:
            try:    self.sdbconn=boto.connect_sdb()
            except: raise SDB_Connect()        
            
    def set_domain(self, domain_name):
        """
        Set the domain on which to work with
        
        @raise SDB_CreateDomain: 
        """
        self.dname=domain_name
        
        try:        self.sdb=self.sdbconn.create_domain(self.dname)
        except:
            try:    self.sdb=self.sdbconn.create_domain(self.dname)
            except: raise SDB_CreateDomain("Can't create SDB domain '%s'" % self.dname)
            
    def delete_domain(self):
        """
        Delete domain  (silent)
        """
        try:
            self.sdbconn.delete_domain(self.dname)
        except:
            pass
            
    def get(self, key, category=None, last=True, limit=20, consistent_read=False):
        """
        Get a specific from, optionally, a given category
        
        @raise SDB_Access:
        """
        if category is None:
            stm="SELECT * from %s where key='%s' and creation_date is not null order by creation_date DESC limit %s" % (self.dname, key, limit)
        else:
            stm="SELECT * from %s where key='%s' and category='%s' and creation_date is not null order by creation_date DESC limit %s" % (self.dname, key, category, limit)
            
        try:                
            rs=self.sdb.select(stm, max_items=limit, consistent_read=consistent_read)
        except:
            raise SDB_Access()

        if last:
            try:    return rs.next()
            except: return None

        return rs
        
    def get_by_category(self, category, limit=20, consistent_read=False, last=True):
        """
        Get a batch of items from a given category
        
        :param category: string
        
        @raise SDB_Access: 
        """
        stm="SELECT * from %s where category='%s' and creation_date is not null order by creation_date DESC limit %s" % (self.dname, category, limit)
        try:
            rs=self.sdb.select(stm, max_items=limit, consistent_read=consistent_read)
        except:
            raise SDB_Access()
        
        if last:
            try:    return rs.next()
            except: return None
        
        return rs

    def delete_item(self, item, silent=True):
        try:
            self.sdb.delete_item(item)
        except:
            if not silent:
                raise SDB_Access()

    def delete(self, key):
        """
        Must find all items associated with 'key' and delete 1 by 1
        
        @return Number of items deleted
        @raise SDB_Access: 
        """
        try:
            s=self.get(key, last=False, consistent_read=True)
            if s is None:
                return 0
            count=0
            for item in s:
                self.sdb.delete_attributes(item.name)
                count=count+1
                return count            
        except Exception, e:
            raise SDB_Access(e)
           
    def del_by_value(self, value, category=None, consistent_read=True, silent=True):
        """
        Deletes entries by 'value' & optionally 'category'
        
        @raise SDB_Access: 
        """
        if category is None:
            stm="SELECT * from %s where value='%s' and creation_date is not null order by creation_date DESC" % (self.dname, value)
        else:
            stm="SELECT * from %s where value='%s' and category='%s' and creation_date is not null order by creation_date DESC" % (self.dname, value, category)
            
        try:                
            rs=self.sdb.select(stm, consistent_read=consistent_read)
        except:
            if silent:
                return 0
            else:
                raise SDB_Access()
            
        try:
            count=0
            for item in rs:
                self.delete_item(item)
                count=count+1
            
            return count
        except:
            if not silent:
                raise SDB_Access()
            return 0
        
            
    def del_by_key_value(self, key, value, category=None, silent=True):
        """
        Deletes a specific entry in the journal
        
        @return # deleted
        @raise SDB_Access: 
        """
        if category is None:
            stm="SELECT * from %s where key='%s' and value='%s' and creation_date is not null order by creation_date DESC" % (self.dname, key, value)
        else:
            stm="SELECT * from %s where key='%s' and value='%s' and category='%s' and creation_date is not null order by creation_date DESC" % (self.dname, key, value, category)
            
        try:                
            rs=self.sdb.select(stm)
            count=0
            for item in rs:
                self.delete_item(item)
                count=count+1
            
            return count            
        except:
            if silent:
                return 0
            else:
                raise SDB_Access()

            
    def insert(self, key, value, attrs=None, category=None, id=None, days=0, hours=0, minutes=0):
        """
        Inserts a [category:key:value] with an optional expiry_date
        
        @return (id, creation iso time, expiration iso time)        
        """
        id=id or uuid.uuid1()

        try:     
            item = self.sdb.new_item(id)
        except:  
            raise SDB_CreateItem("Can't create SDB key(%s) in '%s' domain" % (key, self.dname))
        
        now, then=iso_now_and_then(days=days, hours=hours, minutes=minutes)
        
        item["creation_date"]=now
        item["key"]=key
        item["value"]=value
        if category is not None:
            item["category"]=category
        if then is not None:
            item["expiry_date"]=then
        
        ## Can't use 'extend' as the object 'item' isn't a real dict
        if attrs is not None:
            for k,v in attrs.iteritems():
                item[k]=v
        
        try:    item.save()
        except Exception,e: 
            raise SDB_Access("Can't save item with SDB key(%s) in domain(%s): %s" % (key, self.dname, e))
        
        return (str(id), now, then)

    def get_batch_by_value(self, value, category=None, limit=1000, consistent_read=False):
        """
        Returns a 'batch' of entries based on category & value
        Oldest first
        """
        if category is None:
            stm="SELECT * from %s where value='%s' and creation_date is not null order by creation_date asc limit %s" % (self.dname, value, limit)
        else:
            stm="SELECT * from %s where category='%s' and value='%s' and creation_date is not null order by creation_date asc limit %s" % (self.dname, category, value, limit)
        try:
            rs=self.sdb.select(stm, max_items=limit, consistent_read=consistent_read)
            return rs
        except:
            raise SDB_Access()


    def get_batch(self, limit=1000, by_expiry_date=False, consistent_read=False):
        """
        Returns a 'batch' based on ASC order
        
        NOTE: Legacy... use the more versatile :get_by_batch
        """
        if by_expiry_date:
            stm="SELECT * from %s where expiry_date is not null order by expiry_date ASC limit %s" % (self.dname, limit)
        else:
            stm="SELECT * from %s where creation_date is not null order by creation_date asc limit %s" % (self.dname, limit)
        try:
            rs=self.sdb.select(stm, consistent_read=consistent_read, max_items=limit)
        except:
            raise SDB_Access()
        
        return rs

    def get_by_batch(self, start=0, limit=1000, dir='asc', consistent_read=False):
        """
        Retrieves a batch of record based on 'creation_date'
        """
        #logging.info("kv.get_by_batch: start(%s) limit(%s)" % (start, limit))
        stm="select * from %s where creation_date>'%s' order by creation_date %s limit %s" % (self.dname, start, dir, limit)
        try:
            rs=self.sdb.select(stm, consistent_read=consistent_read, max_items=limit)
        except:
            raise SDB_Access()
        
        return rs


